// Copyright 2019 Citra Emulator Project
// Licensed under GPLv2 or any later version
// Refer to the license.txt file included.
#include "common/logging/log.h"
#include "wmf_decoder_utils.h"

// utility functions
void ReportError(std::string msg, HRESULT hr) {
    if (SUCCEEDED(hr)) {
        return;
    }
    LPSTR err;
    FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_ALLOCATE_BUFFER |
                      FORMAT_MESSAGE_IGNORE_INSERTS,
                  nullptr, hr,
                  // hardcode to use en_US because if any user had problems with this
                  // we can help them w/o translating anything
                  MAKELANGID(LANG_ENGLISH, SUBLANG_ENGLISH_US), (LPSTR)&err, 0, nullptr);
    if (err != nullptr) {
        LOG_CRITICAL(Audio_DSP, "{}: {}", msg, err);
    }
    LOG_CRITICAL(Audio_DSP, "{}: {:08x}", msg, hr);
}

bool MFCoInit() {
    HRESULT hr = S_OK;

    // lite startup is faster and all what we need is included
    hr = MFStartup(MF_VERSION, MFSTARTUP_LITE);
    if (hr != S_OK) {
        // Do you know you can't initialize MF in test mode or safe mode?
        ReportError("Failed to initialize Media Foundation", hr);
        return false;
    }

    LOG_INFO(Audio_DSP, "Media Foundation activated");

    return true;
}

bool MFDecoderInit(IMFTransform** transform, GUID audio_format) {
    HRESULT hr = S_OK;
    MFT_REGISTER_TYPE_INFO reg = {0};
    GUID category = MFT_CATEGORY_AUDIO_DECODER;
    IMFActivate** activate;
    UINT32 num_activate;

    reg.guidMajorType = MFMediaType_Audio;
    reg.guidSubtype = audio_format;

    hr = MFTEnumEx(category,
                   MFT_ENUM_FLAG_SYNCMFT | MFT_ENUM_FLAG_LOCALMFT | MFT_ENUM_FLAG_SORTANDFILTER,
                   &reg, nullptr, &activate, &num_activate);
    if (FAILED(hr) || num_activate < 1) {
        ReportError("Failed to enumerate decoders", hr);
        CoTaskMemFree(activate);
        return false;
    }
    LOG_INFO(Audio_DSP, "Windows(R) Media Foundation found {} suitable decoder(s)", num_activate);
    for (unsigned int n = 0; n < num_activate; n++) {
        hr = activate[n]->ActivateObject(IID_IMFTransform, (void**)transform);
        if (FAILED(hr))
            *transform = nullptr;
        activate[n]->Release();
    }
    if (*transform == nullptr) {
        ReportError("Failed to initialize MFT", hr);
        CoTaskMemFree(activate);
        return false;
    }
    CoTaskMemFree(activate);
    return true;
}

void MFDeInit(IMFTransform* transform) {
    MFShutdownObject(transform);
    CoUninitialize();
}

unique_mfptr<IMFSample> CreateSample(void* data, DWORD len, DWORD alignment, LONGLONG duration) {
    HRESULT hr = S_OK;
    IMFMediaBuffer* buf_tmp = nullptr;
    unique_mfptr<IMFMediaBuffer> buf;
    IMFSample* sample_tmp = nullptr;
    unique_mfptr<IMFSample> sample;

    hr = MFCreateSample(&sample_tmp);
    if (FAILED(hr)) {
        ReportError("Unable to allocate a sample", hr);
        return nullptr;
    }
    sample.reset(sample_tmp);
    // Yes, the argument for alignment is the actual alignment - 1
    hr = MFCreateAlignedMemoryBuffer(len, alignment - 1, &buf_tmp);
    if (FAILED(hr)) {
        ReportError("Unable to allocate a memory buffer for sample", hr);
        return nullptr;
    }
    buf.reset(buf_tmp);
    if (data) {
        BYTE* buffer;
        // lock the MediaBuffer
        // this is actually not a thread-safe lock
        hr = buf->Lock(&buffer, nullptr, nullptr);
        if (FAILED(hr)) {
            ReportError("Unable to lock down MediaBuffer", hr);
            return nullptr;
        }

        std::memcpy(buffer, data, len);

        buf->SetCurrentLength(len);
        buf->Unlock();
    }

    sample->AddBuffer(buf.get());
    hr = sample->SetSampleDuration(duration);
    if (FAILED(hr)) {
        ReportError("Unable to set sample duration, but continuing anyway", hr);
    }

    return std::move(sample);
}

bool SelectInputMediaType(IMFTransform* transform, int in_stream_id, const ADTSData& adts,
                          UINT8* user_data, UINT32 user_data_len, GUID audio_format) {
    HRESULT hr = S_OK;
    IMFMediaType* t;

    // actually you can get rid of the whole block of searching and filtering mess
    // if you know the exact parameters of your media stream
    hr = MFCreateMediaType(&t);
    if (FAILED(hr)) {
        ReportError("Unable to create an empty MediaType", hr);
        return false;
    }

    // basic definition
    t->SetGUID(MF_MT_MAJOR_TYPE, MFMediaType_Audio);
    t->SetGUID(MF_MT_SUBTYPE, audio_format);

    t->SetUINT32(MF_MT_AAC_PAYLOAD_TYPE, 1);
    t->SetUINT32(MF_MT_AUDIO_NUM_CHANNELS, adts.channels);
    t->SetUINT32(MF_MT_AUDIO_SAMPLES_PER_SECOND, adts.samplerate);
    // 0xfe = 254 = "unspecified"
    t->SetUINT32(MF_MT_AAC_AUDIO_PROFILE_LEVEL_INDICATION, 254);
    t->SetUINT32(MF_MT_AUDIO_BLOCK_ALIGNMENT, 1);
    t->SetBlob(MF_MT_USER_DATA, user_data, user_data_len);
    hr = transform->SetInputType(in_stream_id, t, 0);
    if (FAILED(hr)) {
        ReportError("failed to select input types for MFT", hr);
        return false;
    }

    return true;
}

bool SelectOutputMediaType(IMFTransform* transform, int out_stream_id, GUID audio_format) {
    HRESULT hr = S_OK;
    UINT32 tmp;
    IMFMediaType* type;
    unique_mfptr<IMFMediaType> t;

    // If you know what you need and what you are doing, you can specify the condition instead of
    // searching but it's better to use search since MFT may or may not support your output
    // parameters
    for (DWORD i = 0;; i++) {
        hr = transform->GetOutputAvailableType(out_stream_id, i, &type);
        t.reset(type);
        if (hr == MF_E_NO_MORE_TYPES || hr == E_NOTIMPL) {
            return true;
        }
        if (FAILED(hr)) {
            ReportError("failed to get output types for MFT", hr);
            return false;
        }

        hr = t->GetUINT32(MF_MT_AUDIO_BITS_PER_SAMPLE, &tmp);

        if (FAILED(hr))
            continue;
        // select PCM-16 format
        if (tmp == 32) {
            hr = t->SetUINT32(MF_MT_AUDIO_BLOCK_ALIGNMENT, 1);
            if (FAILED(hr)) {
                ReportError("failed to set MF_MT_AUDIO_BLOCK_ALIGNMENT for MFT on output stream",
                            hr);
                return false;
            }
            hr = transform->SetOutputType(out_stream_id, t.get(), 0);
            if (FAILED(hr)) {
                ReportError("failed to select output types for MFT", hr);
                return false;
            }
            return true;
        } else {
            continue;
        }

        return false;
    }

    ReportError("MFT: Unable to find preferred output format", E_NOTIMPL);
    return false;
}

int DetectMediaType(char* buffer, size_t len, ADTSData* output, char** aac_tag) {
    if (len < 7) {
        return -1;
    }

    ADTSData tmp;
    // see https://docs.microsoft.com/en-us/windows/desktop/api/mmreg/ns-mmreg-heaacwaveinfo_tag
    // for the meaning of the byte array below

    // it might be a good idea to wrap the parameters into a struct
    // and pass that struct into the function but this will lead to messier code
    // const UINT8 aac_data[] = { 0x01, 0x00, 0xfe, 00, 00, 00, 00, 00, 00, 00, 00, 00, 0x11, 0x90
    // }; first byte: 0: raw aac 1: adts 2: adif 3: latm/laos
    UINT8 aac_tmp[] = {0x01, 0x00, 0xfe, 00, 00, 00, 00, 00, 00, 00, 00, 00, 0x00, 0x00};
    uint16_t tag = 0;

    uint32_t result = ParseADTS(buffer, &tmp);
    if (result == 0) {
        return -1;
    }

    tag = MFGetAACTag(tmp);
    aac_tmp[12] |= (tag & 0xff00) >> 8;
    aac_tmp[13] |= (tag & 0x00ff);
    std::memcpy(*aac_tag, aac_tmp, 14);
    std::memcpy(output, &tmp, sizeof(ADTSData));
    return 0;
}

void MFFlush(IMFTransform* transform) {
    HRESULT hr = transform->ProcessMessage(MFT_MESSAGE_COMMAND_FLUSH, 0);
    if (FAILED(hr)) {
        ReportError("MFT: Flush command failed", hr);
    }
    hr = transform->ProcessMessage(MFT_MESSAGE_NOTIFY_END_OF_STREAM, 0);
    if (FAILED(hr)) {
        ReportError("Failed to end streaming for MFT", hr);
    }
}

MFInputState SendSample(IMFTransform* transform, DWORD in_stream_id, IMFSample* in_sample) {
    HRESULT hr = S_OK;

    if (in_sample) {
        hr = transform->ProcessInput(in_stream_id, in_sample, 0);
        if (hr == MF_E_NOTACCEPTING) {
            return NOT_ACCEPTED; // try again
        } else if (FAILED(hr)) {
            ReportError("MFT: Failed to process input", hr);
            return INPUT_ERROR;
        } // FAILED(hr)
    } else {
        hr = transform->ProcessMessage(MFT_MESSAGE_COMMAND_DRAIN, 0);
        if (FAILED(hr)) {
            ReportError("MFT: Failed to drain when processing input", hr);
        }
    }

    return INPUT_OK;
}

std::tuple<MFOutputState, unique_mfptr<IMFSample>> ReceiveSample(IMFTransform* transform,
                                                                 DWORD out_stream_id) {
    HRESULT hr;
    MFT_OUTPUT_DATA_BUFFER out_buffers;
    MFT_OUTPUT_STREAM_INFO out_info;
    DWORD status = 0;
    unique_mfptr<IMFSample> sample;
    bool mft_create_sample = false;

    hr = transform->GetOutputStreamInfo(out_stream_id, &out_info);

    if (FAILED(hr)) {
        ReportError("MFT: Failed to get stream info", hr);
        return std::make_tuple(FATAL_ERROR, std::move(sample));
    }
    mft_create_sample = (out_info.dwFlags & MFT_OUTPUT_STREAM_PROVIDES_SAMPLES) ||
                        (out_info.dwFlags & MFT_OUTPUT_STREAM_CAN_PROVIDE_SAMPLES);

    while (true) {
        status = 0;

        if (!mft_create_sample) {
            sample = CreateSample(nullptr, out_info.cbSize, out_info.cbAlignment);
            if (!sample.get()) {
                ReportError("MFT: Unable to allocate memory for samples", hr);
                return std::make_tuple(FATAL_ERROR, std::move(sample));
            }
        }

        out_buffers.dwStreamID = out_stream_id;
        out_buffers.pSample = sample.get();

        hr = transform->ProcessOutput(0, 1, &out_buffers, &status);

        if (!FAILED(hr)) {
            break;
        }

        if (hr == MF_E_TRANSFORM_NEED_MORE_INPUT) {
            // Most likely reasons: data corrupted; your actions not expected by MFT
            return std::make_tuple(NEED_MORE_INPUT, std::move(sample));
        }

        if (hr == MF_E_TRANSFORM_STREAM_CHANGE) {
            ReportError("MFT: stream format changed, re-configuration required", hr);
            return std::make_tuple(NEED_RECONFIG, std::move(sample));
        }

        break;
    }

    if (out_buffers.dwStatus & MFT_OUTPUT_DATA_BUFFER_INCOMPLETE) {
        // this status is also unreliable but whatever
        return std::make_tuple(HAVE_MORE_DATA, std::move(sample));
    }

    if (out_buffers.pSample == nullptr) {
        ReportError("MFT: decoding failure", hr);
        return std::make_tuple(FATAL_ERROR, std::move(sample));
    }

    return std::make_tuple(OK, std::move(sample));
}

int CopySampleToBuffer(IMFSample* sample, void** output, DWORD* len) {
    unique_mfptr<IMFMediaBuffer> buffer;
    IMFMediaBuffer* tmp;
    HRESULT hr = S_OK;
    BYTE* data;

    hr = sample->GetTotalLength(len);
    if (FAILED(hr)) {
        ReportError("Failed to get the length of sample buffer", hr);
        return -1;
    }

    hr = sample->ConvertToContiguousBuffer(&tmp);
    if (FAILED(hr)) {
        ReportError("Failed to get sample buffer", hr);
        return -1;
    }
    buffer.reset(tmp);

    hr = tmp->Lock(&data, nullptr, nullptr);
    if (FAILED(hr)) {
        ReportError("Failed to lock the buffer", hr);
        return -1;
    }

    *output = malloc(*len);
    std::memcpy(*output, data, *len);

    // if buffer unlock fails, then... whatever, we have already got data
    buffer->Unlock();
    return 0;
}
